/* $Id: ncron.c 9 2023-02-22 03:06:23Z nishi $ */

#include <ncron.h>

#include <ncron/parse.h>
#include <ncron/util.h>

#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <time.h>
#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/wait.h>
#include <sys/stat.h>

#define NCRON_MINUTE	1 << 0
#define NCRON_HOUR	1 << 1
#define NCRON_DATE	1 << 2
#define NCRON_MONTH	1 << 3
#define NCRON_DAY	1 << 4

#define WRITE(fd,str) write(fd, str, sstrlen(str))

extern char** argv;
extern int argc;

void logreset(int logfd){
	fclose(fopen("/var/log/cron", "w"));
	WRITE(logfd, "===== NCron v");
	WRITE(logfd, NCRON_VERSION);
	WRITE(logfd, "\n");
	WRITE(logfd, "===== Log file turned over\n");
}

bool ncron_match(struct ncron_entry* entry, struct tm* tm){
	if(entry->e_reboot == 1) return false;
	int i;
	char fullflag = NCRON_MINUTE | NCRON_HOUR | NCRON_DATE | NCRON_MONTH | NCRON_DAY;
	for(i = 0; entry->e_hour_start[i] != -1; i++){
		char flag = 0;
		if(((entry->e_minute_start[i] <= tm->tm_min) && (entry->e_minute_end[i] == -2 ? (entry->e_minute_start[i] == tm->tm_min) : (tm->tm_min <= entry->e_minute_end[i]))) && ((tm->tm_min % entry->e_minute_step[i]) == 0)) flag |= NCRON_MINUTE;
		if(((entry->e_hour_start[i] <= tm->tm_hour) && (entry->e_hour_end[i] == -2 ? (entry->e_hour_start[i] == tm->tm_hour) : (tm->tm_hour <= entry->e_hour_end[i]))) && ((tm->tm_hour % entry->e_hour_step[i]) == 0)) flag |= NCRON_HOUR;
		if(((entry->e_date_start[i] <= tm->tm_mday) && (entry->e_date_end[i] == -2 ? (entry->e_date_start[i] == tm->tm_mday) : (tm->tm_mday <= entry->e_date_end[i]))) && ((tm->tm_mday % entry->e_date_step[i]) == 0)) flag |= NCRON_DATE;
		if(((entry->e_month_start[i] <= tm->tm_mon) && (entry->e_month_end[i] == -2 ? (entry->e_month_start[i] == tm->tm_mon) : (tm->tm_mon <= entry->e_month_end[i]))) && ((tm->tm_mon % entry->e_month_step[i]) == 0)) flag |= NCRON_MONTH;
		if(((entry->e_day_start[i] <= tm->tm_wday) && (entry->e_day_end[i] == -2 ? (entry->e_day_start[i] == tm->tm_wday) : (tm->tm_wday <= entry->e_day_end[i]))) && ((tm->tm_wday % entry->e_day_step[i]) == 0)) flag |= NCRON_DAY;
		if(flag == fullflag) return true;
	}
	return false;
}

char* ncron_date(){
	char* date = malloc(128);
	time_t t = time(NULL);
	struct tm *time_st = localtime(&t);
	strftime(date, 128, "%a %b %e %H:%M:%S %Z %Y", time_st);
	return date;
}

void ncron_start(){
	signal(SIGCHLD, SIG_IGN);
	int pid;
	struct ncrontab* tab = ncron_parse();
	int i;
	int nullfd = open("/dev/null", O_RDWR);
	int logfd = open("/var/log/cron", O_RDWR | O_CREAT | O_TRUNC);
	close(logfd);
	logfd = fileno(stdout);
	if(logfd == -1) exit(-1);
	WRITE(logfd, "===== NCron v");
	WRITE(logfd, NCRON_VERSION);
	WRITE(logfd, "\n");
	int reboot_pid = fork();
	if(reboot_pid == 0){
		dup2(logfd, 1);
		dup2(logfd, 2);
		fflush(stdout);
		for(i = 0; tab->tabs[i] != NULL; i++){
			if(fork() != 0) continue;
			int* fd = malloc(sizeof(*fd) * 2);
			pipe(fd);
			int job_pid = fork();
			if(job_pid == 0){
				if(tab->tabs[i]->e_reboot){
					close(fd[0]);
					dup2(nullfd, 1);
					dup2(1, 2);
					char** env = malloc(sizeof(*env) * 2);
					env[0] = sstrcat("PATH=", tab->tabs[i]->e_path);
					env[1] = NULL;
					if(0 != setgid(tab->tabs[i]->e_gid)){
						WRITE(fd[1], " (ERROR) setgid() failed");
						_exit(0);
					}
					if(0 != setuid(tab->tabs[i]->e_uid)){
						WRITE(fd[1], " (ERROR) setuid() failed");
						_exit(0);
					}
					execle(tab->tabs[i]->e_shell, tab->tabs[i]->e_shell, "-c", tab->tabs[i]->e_command, NULL, env);
				}
				_exit(-1);
			}else{
				if(tab->tabs[i]->e_reboot){
					close(fd[1]);
					char* date;
					date = ncron_date();
					printf("%s: @reboot: PID (%d) START\n", date, job_pid);
					free(date);
					fflush(stdout);
					argv[0] = sstrdup("ncron: Job");
					argv[1] = NULL;
					waitpid(job_pid, 0, 0);
					date = ncron_date();
					char* log = malloc(128);
					memset(log, 0, 128);
					read(fd[0], log, 128);
					printf("%s: @reboot: PID (%d) END%s\n", date, job_pid, log);
					free(date);
					fflush(stdout);
				}
			}
			free(fd);
			_exit(0);
		}
		_exit(0);
	}
#ifndef SIGNAL_LOGRESET
	if(fork() == 0){
		argv[0] = sstrdup("ncron: Log Size Watcher");
		argv[1] = NULL;
		while(1){
			if(getppid() == 1) break;
			struct stat s;
			stat("/var/log/cron.log", &s);
			if(s.st_size > 1 * 1024 * 1024){
				logreset(logfd);
			}
			sleep(1);
		}
		_exit(0);
	}
#else
	logreset(logfd);
#endif
	argv[0] = sstrdup("ncron: Main Loop");
	argv[1] = NULL;
	time_t now = time(0);
	struct tm* tm = localtime(&now);
	/*sleep(60 - (tm->tm_sec == 0 ? 60 : tm->tm_sec));	* Wait until the next minute... */
	while(1){
		now = time(0);
		tm = localtime(&now);
	       	pid = fork();
		if(pid == 0){
			dup2(logfd, 1);
			dup2(logfd, 2);
			for(i = 0; tab->tabs[i] != NULL; i++){
				if(fork() != 0) continue;
				bool matched = ncron_match(tab->tabs[i], tm);
				int* fd = malloc(sizeof(*fd) * 2);
				pipe(fd);
				int job_pid = fork();
				if(job_pid == 0){
					if(!matched) _exit(-1);
					if(matched){
						close(fd[0]);
						dup2(nullfd, 1);
						dup2(1, 2);
						char** env = malloc(sizeof(*env) * 2);
						env[0] = sstrcat("PATH=", tab->tabs[i]->e_path);
						env[1] = NULL;
						if(0 != setgid(tab->tabs[i]->e_gid)){
							WRITE(fd[1], " (ERROR) setgid() failed");
							_exit(0);
						}
						if(0 != setuid(tab->tabs[i]->e_uid)){
							WRITE(fd[1], " (ERROR) setuid() failed");
							_exit(0);
						}
						execle(tab->tabs[i]->e_shell, tab->tabs[i]->e_shell, "-c", tab->tabs[i]->e_command, NULL, env);
					}
					_exit(-1);
				}else{
					if(matched){
						close(fd[1]);
						char* date;
						date = ncron_date();
						printf("%s: PID (%d) START (entry in %s)\n", date, job_pid, tab->tabs[i]->e_crontab);
						free(date);
						fflush(stdout);
						argv[0] = sstrdup("ncron: Job");
						argv[1] = NULL;
						waitpid(job_pid, 0, 0);
						date = ncron_date();
						char* log = malloc(128);
						memset(log, 0, 128);
						read(fd[0], log, 128);
						printf("%s: PID (%d) END%s\n", date, job_pid, log);
						free(date);
						fflush(stdout);
					}
				}
				free(fd);
				_exit(0);
			}
			_exit(0);
		}else{
			sleep(60);
		}
	}
	ncron_ncrontab_free(tab);
}
